#!/bin/python3
#%%
import io, os, re
import openpyxl
from openpyxl.styles import Font,Alignment,Color,Border,Side,PatternFill
from openpyxl.drawing.image import Image
from openpyxl.formatting.rule import DataBarRule, ColorScaleRule
from openpyxl.utils import get_column_letter, column_index_from_string
from openpyxl.drawing.spreadsheet_drawing import TwoCellAnchor, OneCellAnchor, AnchorMarker
from openpyxl.utils.units import pixels_to_EMU

import pandas as pd

#%%

color_dict={
        'H1':['B0C4DE',Font(bold=True),8],
        'H2':['FFA07A',Font(),9],
        'H3':['90EE90',Font(italic=True),10],
    }

def cell_set_bold(cell: openpyxl.cell.cell.Cell):
    cell.font = Font(bold=True)
    return None

def cell_set_border(cell: openpyxl.cell.cell.Cell):
    thin = Side(border_style="thin", color="000000")
    border = Border(left=thin, 
                right=thin, 
                top=thin, 
                bottom=thin)
    cell.border = border
    return None

def cell_set_number_format(cell: openpyxl.cell.cell.Cell, format = '0.000'):
    cell.number_format = format
    return None


def cell_set_alignment(cell: openpyxl.cell.cell.Cell, horizontal='center', vertical='center'):
    cell.alignment = Alignment(horizontal=horizontal, vertical=vertical)
    return None


def sheet_set_dimensions(worksheet: openpyxl.worksheet.worksheet.Worksheet, column = None, column_width = None, index = None, index_height = None):
    if isinstance(column_width, str) and column_width.endswith('px'):
        column_width = float(column_width[:-2]) * .125
    if isinstance(index_height, str) and index_height.endswith('px'):
        index_height = float(index_height[:-2]) * .75
    if column is not None:
        worksheet.column_dimensions[column].width = column_width
    elif index is not None:
        worksheet.row_dimensions[index].height = index_height
    return None


def resize_img(width, height, max_size, ratio = 1):
    if not isinstance(max_size, list):
        max_size = [max_size] * 2
    max_width, max_height = max_size
    print(f'old: width/height = {width}/{height} ({width/height})')
    new_height = max_width / width * height
    new_width = max_height / height * width
    if new_height <= max_height:
        width, height = max_width, new_height
    elif new_width <= max_width:
        width, height = new_width, max_height
    else:
        raise ValueError
    width *= ratio
    height *= ratio
    print(f'new: width/height = {width}/{height} ({width/height})')
    return width, height


def split_string(s):
    match = re.match(r"([a-zA-Z]+)(\d+)", s)
    if match:
        return match.groups()
    else:
        return None


def column_to_letters(worksheet: openpyxl.worksheet.worksheet.Worksheet, columns = [], row_num = 1, col_num = 10000):
    if isinstance(columns, str):
        column = columns
        column_idx = None
        for col in range(1, col_num+1):
            if worksheet.cell(row=row_num, column=col).value == column:
                column_idx = col
                break
        if col == col_num:
            column_idx = None
        else:
            column_idx = get_column_letter(column_idx)
    elif isinstance(columns, list):
        column_idx = [column_to_letters(worksheet=worksheet, columns=column, row_num=row_num) for column in columns]
    return column_idx

def all_column_to_letters(worksheet: openpyxl.worksheet.worksheet.Worksheet, col_num = 1, row_num = 1):
    columns = []
    for col in range(1,col_num+2):
        columns.append(worksheet.cell(row=row_num, column=col).value) 
    column_idx = [column_to_letters(worksheet=worksheet, columns=column, row_num=row_num) for column in columns]
    return column_idx

def insert_dataframes_to_worksheet(worksheet: openpyxl.worksheet.worksheet.Worksheet, df: pd.DataFrame, df_name: str, index_and_columns = True, start = 1,head = None,df_info_ori={}):
    print(head)
    n_rows, n_cols = df.shape
    

    fill_blue = PatternFill(start_color="add8e6",
                        end_color="add8e6",
                        fill_type="solid")
    
    if head is not None:
        tmp = start-len(head)+1
        if 'null' not in head:
            for level in head:
                worksheet.cell(tmp, 2).value = head[level]
                worksheet.cell(tmp, 2).font = color_dict[level][1]
                for i in range(1,color_dict[level][2]):
                    worksheet.cell(tmp,i+1).fill = PatternFill(start_color=color_dict[level][0],
                                    end_color=color_dict[level][0],
                                    fill_type="solid")
                tmp=tmp+1
    else:
        for i in range(1,6):
            worksheet.cell(start,i).fill = fill_blue

    start += 1
    if index_and_columns:

        gap = 2

        if df.index.name is None:
            index_name = 'index'
        else:
            index_name = df.index.name
        worksheet.cell(start, 1).value = index_name

        for r in range(n_rows):
            worksheet.cell(r+1+start, 1).value = df.index.to_series().iloc[r]
            del r
        
        head_name_color = df_info_ori.get('head_color','87CEEB')
        tail_name_color = df_info_ori.get('tail_color','')
        
        for c in range(n_cols):
            worksheet.cell(start, c+2).value = df.columns.to_series().iloc[c]
            if head_name_color!='':
                worksheet.cell(start, c+2).fill = PatternFill(start_color=head_name_color,end_color=head_name_color,fill_type="solid")
            if tail_name_color!='':
                worksheet.cell(start+n_rows,c+2).fill = PatternFill(start_color=tail_name_color,end_color=tail_name_color,fill_type="solid")
            del c
    
    else:
        gap = 1

    for r in range(n_rows):
        for c in range(n_cols):
            worksheet.cell(r+gap+start-1, c+gap).value = df.iloc[r, c]
            
    return worksheet


def sheet_to_dataframe(sheet: openpyxl.worksheet.worksheet.Worksheet, index_col = 'auto', header = True, df_cell=[]):
    m,n,row_num = df_cell[0],df_cell[1],df_cell[2]
    start_row, start_col = row_num, 1  
    end_row, end_col = row_num+m, 1+n 
    
    data = sheet.iter_rows(min_row=start_row, min_col=start_col, max_row=end_row, max_col=end_col,values_only =True)
    if header:
        df = pd.DataFrame(columns=next(data))
    else:
        df = pd.DataFrame(columns=[f'{k}' for k in range(sheet.max_column)])
    k = 0

    while True:
        try:
            df.loc[k] = next(data)
            k += 1
        except StopIteration :
            break
    if index_col == 'auto':
        index_col = 'index'
    if index_col is not None:
        df.set_index(keys=index_col, inplace=True)
    return df


def fix_columns_of_data_rule(columns: list):
    if any([isinstance(k, str) for k in columns]) and any([isinstance(k, list) for k in columns]):
        columns = [[k] if isinstance(k, str) else k for k in columns]
    return columns


def flatten_columns(columns: list):
    if isinstance(columns, str):
        columns = [columns]
    new_columns = []
    if all([isinstance(k, str) for k in columns]):
        new_columns = sorted(set(columns))
        print('columns =', columns)
    else:
        for a in columns:
            if isinstance(a, str):
                new_columns.append(a)
            elif isinstance(a, list):
                new_columns.extend(a)
            else:
                raise ValueError(f'found a = {a} in columns.')
        new_columns = flatten_columns(new_columns)
    return new_columns

#%%

class SheetModifier():

    def __init__(self, worksheet, dfs_info , row1_as_col = False) -> None:
        self.worksheet = worksheet
        self.row1_as_col = row1_as_col
        self.dfs_info = dfs_info
        self.df_name = None
        return None
    
    def __repr__(self) -> str:
        return str(self.to_frame())
    
    def __str__(self) -> str:
        return str(self.__repr__())
    
    def __getitem__(self, df_name):
        print(f'df_name : {df_name}')
        self.df_name = df_name
        ret = self
        return ret

    def set_bold_sereis(self, worksheet, column = None, index = None, start = 0, end = 2):
        if column is not None:
            for cell in worksheet[column][start:end]:
                cell_set_bold(cell=cell)
        if index is not None:
            for cell in worksheet[index][start:end]:
                cell_set_bold(cell=cell)
        return None
    
    def set_border_sereis(self, worksheet,column = None, index = None, start = 0, end = 2):
        if column is not None:
            for cell in worksheet[column][start:end]:
                cell_set_border(cell=cell)
        if index is not None:
            for cell in worksheet[index][start:end]:
                cell_set_border(cell=cell)
        return None
    
    def set_sheet_bold(self, columns = [], index_is = True):
        print('set_sheet_bold()')
        df_name = self.df_name
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name =self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    else:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet=worksheet, columns=columns_need, row_num=row_num)
                    for column in columns_need:
                        self.set_bold_sereis(worksheet=worksheet,column=column, start=start-1, end=start)
                    if index_is:
                        self.set_bold_sereis(worksheet=worksheet,column='A',start=start,end=end-1)
        return None
    
    def set_sheet_border(self, columns = [], index_is =True, all_col = True):
        print('set_sheet_border()')
        df_name = self.df_name
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name =self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    else:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet=worksheet, columns=columns_need, row_num=row_num)
                    for column in columns_need:
                        if all_col:
                            self.set_border_sereis(worksheet=worksheet,column=column, start=start-1, end=end-1)
                        else:
                            self.set_border_sereis(worksheet=worksheet,column=column, start=start-1, end=start)
                    if index_is:
                        self.set_border_sereis(worksheet=worksheet,column='A',start=start,end=end-1)
        return None


    def set_cols_bold(self, columns = [], start = 0):
        print('set_cols_bold()')
        df_name = self.df_name
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name =self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    elif self.row1_as_col:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet, columns=columns_need, row_num=row_num)
                    for column in columns_need:
                        self.set_bold_sereis(worksheet=worksheet,column=column, start=start, end=end)
        return None

    def set_hidden_sereis(self, worksheet,column = None, index = None, start = 0):
        if column is not None:
            worksheet.column_dimensions[column].hidden = True
        if index is not None:
            worksheet.row_dimensions[index].hidden = True
        return None
    
    def set_cols_hidden(self, columns = [], start = 0):
        print('set_cols_hidden()')
        columns = flatten_columns(columns=columns)
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                df_info = dfs_info[df_na]
                m,n,row_num = df_info[0],df_info[1],df_info[2]
                start,end = row_num,row_num+m+1
                if self.row1_as_col:
                    columns_need = column_to_letters(worksheet, columns=columns, row_num=row_num)
                print(f'sheetname : {sheetname} \n df_name : {df_na} columns : {columns_need}')
                for column in columns_need:
                    self.set_hidden_sereis(worksheet=worksheet,column=column, start=start)
        return None

    def set_number_format(self, worksheet,column = None, index = None, format = '0.000', start = 0, end = 2):
        if column is not None:
            for cell in worksheet[column][start:end]:
                cell_set_number_format(cell=cell, format=format)
        if index is not None:
            for cell in worksheet[index][start:end]:
                cell_set_number_format(cell=cell, format=format)
        return None
    
    def set_cols_number_format(self, columns = [], format = '0.000'):
        print(f'set_cols_number_format()')
        columns = flatten_columns(columns=columns)
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col:
                        columns_need = column_to_letters(worksheet, columns=columns, row_num=row_num)
                    print(f'sheetname : {sheetname} \n df_name : {df_na} columns : {columns_need}')
                    for column in columns_need:
                        self.set_number_format(worksheet=worksheet,column=column, format=format, start=start,end=end)
        return None
    
    def set_alignment(self, worksheet,column = None, index = None, horizontal='center', vertical='center', start = 0, end = 1):
        if column is not None:
            for cell in worksheet[column][start:end]:
                cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
        if index is not None:
            for cell in worksheet[index][start:end]:
                cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
        return None
    
    def set_cols_alignment(self, columns = [], horizontal='center', vertical='center', start = 0, name_is = True, all_col = False):
        print('set_cols_alignment()')
        df_name = self.df_name
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name =self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num,d = df_info[0],df_info[1],df_info[2],df_info[3]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    elif self.row1_as_col:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet, columns=columns_need, row_num=row_num)
                    for column in columns_need:
                        if name_is:
                            self.set_alignment(worksheet=worksheet,column=column,horizontal=horizontal, vertical=vertical, start=start-1,end=end)
                        else:
                            # print([column,start,end])
                            self.set_alignment(worksheet=worksheet,column=column,horizontal=horizontal, vertical=vertical, start=start,end=end)
        return None
    
    def set_sheet_alignment(self, columns = [], horizontal='center', vertical='center',index_is =True,name_is = True):
        print('set_sheet_alignment()')
        df_name = self.df_name
        for sheetname in self.worksheet:
            print(sheetname)
            worksheet = self.worksheet[sheetname]
            df_name =self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num,d = df_info[0],df_info[1],df_info[2],df_info[3]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    else:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet=worksheet, columns=columns_need, row_num=row_num)
                    for column in columns_need:
                        for row in range(start-1,start):
                            cell = worksheet[column][row]
                            cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
                    if index_is:
                        for row in range(start-1,end-1):
                            cell = worksheet['A'][row]
                            cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
                    if name_is:
                        for i in range(start-d-1,start-1):
                            cell = worksheet['B'][i]
                            cell_set_alignment(cell=cell, horizontal = 'left', vertical=vertical)
        return None
    
        

    def set_dimensions(self, worksheet,column = None, column_width = None, index = None, index_height = None):
        sheet_set_dimensions(worksheet=worksheet, column = column, column_width = column_width, index = index, index_height = index_height)
        return None
    
    def set_cols_dimensions(self, columns = [], column_width = None, all_col=False):
        print('set_cols_dimensions()')
        columns = flatten_columns(columns=columns)
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and all_col:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                        # columns_need.append('B')
                    else:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet=worksheet, columns=columns_need, row_num=row_num)
                    # print(columns_need)
                    for column in columns_need:
                        self.set_dimensions(worksheet=worksheet,column = column, column_width = column_width)
        return None
    
    def set_indexes_dimensions(self, indexes = None, index_height = None, start = 2,is_name = True):
        print('set_indexes_dimensions()')
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                df_info = dfs_info[df_na]
                m,n,row_num = df_info[0],df_info[1],df_info[2]
                start,end = row_num,row_num+m+1
                print(f'sheetname : {sheetname} \n df_name : {df_na}')
                if is_name:
                    start = start - dfs_info[df_na][3]
                if indexes is None:
                    indexes_need = range(start, end)
                else:
                    indexes_need = indexes + start
                for index in indexes_need:
                    self.set_dimensions(worksheet=worksheet,index = index, index_height = index_height)
        return None
    

    def insert_image(self, worksheet,anchor: str, img = None, resize = 'auto', func = None, ratio = .95):

        if isinstance(img, str) and os.path.isfile(img):
            with open(img, 'rb') as f:
                img = f.read()
        elif img is None and func is not None:
            cell_value = worksheet[anchor].value
            img = func(cell_value)
        if isinstance(img, bytes):
            img = io.BytesIO(img)
            img = Image(img=img)

        flag = 0
        
        if resize == 'auto':
            col, row = split_string(anchor)
            row = int(row)
            print(f'anchor = {anchor} -> {(row, col)}')
            if  col != 'None':
                flag=1
                col_num = column_index_from_string(col)
                cell_width = worksheet.column_dimensions[col].width
                cell_height = worksheet.row_dimensions[row].height
                if cell_width is None:
                    cell_width = 8.38
                if cell_height is None:
                    cell_height = 13.5
                cell_width_pix = cell_width * 8 
                cell_height_pix = cell_height * 4/3 
                print(f'size({(row, col)}) = {(cell_width_pix, cell_height_pix)}')
                img_width, img_height = img.width, img.height
                img_width, img_height = resize_img(img_width, img_height, max_size=[cell_width_pix, cell_height_pix], ratio = ratio)
                
                col_off = pixels_to_EMU((cell_width_pix - img_width) / 2)
                row_off = pixels_to_EMU((cell_height_pix - img_height) / 2)
                _from = AnchorMarker(col_num-1, col_off, row-1, row_off) # dst
                to = AnchorMarker(col_num, -col_off, row, -row_off)
            # cell_width_emu = pixels_to_EMU(cell_width_pix)
            # cell_height_emu = pixels_to_EMU(cell_height_pix)
            # to = AnchorMarker(
            #     col_num-1, cell_width_emu-col_off, 
            #     row-1, cell_height_emu-row_off
            #     )
            # anchor = OneCellAnchor(_from = to)
            # img.anchor = TwoCellAnchor('twoCell', _from, to)
                anchor = TwoCellAnchor('twoCell', _from, to)

        # worksheet.add_image(img=img)
        if flag == 1 and resize == 'auto':
            worksheet.add_image(img=img, anchor=anchor)

        return None


    def insert_col_img(self, column = [], img = None, resize = 'auto', func = None, start = 2, ratio = .95):
        print('insert_col_img()')
        columns = flatten_columns(columns=column)
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num,row_num+m+1
                    if self.row1_as_col and columns == []:
                        columns_need = all_column_to_letters(worksheet=worksheet, col_num=n, row_num=row_num)
                    else:
                        columns_need = flatten_columns(columns=columns)
                        columns_need = column_to_letters(worksheet=worksheet, columns=columns_need, row_num=row_num)
                    print(f'sheetname : {sheetname} \n df_name : {df_na} columns : {columns_need}')
                    for column in columns_need:
                        for k in range(start+1,end):
                            self.insert_image(worksheet=worksheet,anchor = f'{column}{k}', img = img, resize = resize, func = func, ratio=ratio)
        return None
    

    def add_data_rule(self, worksheet,column: str, value_range = 'min-max', rule = 'ColorScale', start = 2,end = 3):
        

        if rule == 'DataBar':
            rule = DataBarRule(
                start_type='num', 
                start_value=min(value_range), 
                end_type='num', 
                end_value=max(value_range), 
                color=Color('FF638EC6')
                )
        elif rule == 'ColorScale':
            rule = ColorScaleRule(
                start_value=min(value_range),
                start_type='num',
                # start_type='min',
                start_color=Color(rgb='63BE7B'),  # green
                mid_value=value_range[1] if len(value_range) == 3 else (min(value_range)+max(value_range))/2,
                mid_type='num',
                # mid_type='percentile',
                mid_color=Color(rgb='FFEB84'), # yellow
                end_value=max(value_range),
                end_type='num',
                # end_type='max',
                end_color=Color(rgb='F8696B')    # red
                )
        worksheet.conditional_formatting.add(f'{column}{start}:{column}{end}', rule)

        return None
    
    
    def add_cols_data_rule(self, columns: str, value_range = 'min-max', rule = 'ColorScale', start = 2):
        print('add_cols_data_rule()')
        columns = flatten_columns(columns=columns)
        for sheetname in self.worksheet:
            worksheet = self.worksheet[sheetname]
            df_name = self.df_name
            dfs_info = self.dfs_info[sheetname]
            if isinstance(df_name,str):
                df_name = [df_name]
            elif isinstance(df_name,slice):
                df_name = [df_na for df_na in dfs_info]
            for df_na in df_name:
                if df_na in dfs_info:
                    df_info = dfs_info[df_na]
                    m,n,row_num = df_info[0],df_info[1],df_info[2]
                    start,end = row_num+1,row_num+m
                    columns_need = fix_columns_of_data_rule(columns=columns)
                    flag = all([isinstance(k, str) for k in columns_need])
                    if flag:
                        if self.row1_as_col:
                            sheet_df = sheet_to_dataframe(worksheet,df_cell=[m,n,row_num])
                            for column in columns_need:
                                if column in sheet_df:
                                    col_df = sheet_df[column]
                                    col_min = col_df.min().min()
                                    col_max = col_df.max().max()
                                    if isinstance(value_range, str) and value_range == 'min-max':
                                        value_range_tmp = [col_min, col_max]
                                        value_range_tmp = [float(x) for x in value_range_tmp]
                                    else:
                                        value_range_tmp = value_range
                                    print(f'columns = {column}, value_range = \'min-max\' -> {value_range_tmp}')
                                    column = column_to_letters(worksheet, columns=column,row_num=row_num)
                                    self.add_data_rule(worksheet=worksheet,column = column, value_range = value_range_tmp, rule = rule, start = start, end = end)
                    else:
                        for column in columns_need:
                            self.add_cols_data_rule(columns=column, value_range=value_range, rule=rule, start=start)
            
        return None
    

    def set_filter(self, freeze_panes = True):
        worksheet = self.worksheet
        max_row = worksheet.max_row
        max_column_letter = get_column_letter(worksheet.max_column)
        worksheet.auto_filter.ref = f"A1:{max_column_letter}{max_row}"
        if freeze_panes:
            worksheet.freeze_panes = 'A2'
        return None
        

def parse_titles(title_str: str):
    title_dict = {}
    if isinstance(title_str, dict):
        title_dict = title_str
    elif isinstance(title_str, str):
        if 'H1' not in title_str and 'H2' not in title_str and 'H3' not in title_str:
            title_dict['H1'] = title_str
        else:
            d = [[k.strip() for k in line.split(' ') if k] for line in title_str.split('\n') if line.strip() != '']
            title_dict = {k[0]: ' '.join(k[1:]) for k in d}
    return title_dict

#%%

class ExcelModifier():

    def __init__(self, xlsx_path = None, workbook = None, df = None, dfs = None, row1_as_col = False) -> None:

        if isinstance(workbook, openpyxl.Workbook):
            pass

        elif isinstance(xlsx_path, str):
            workbook = openpyxl.load_workbook(xlsx_path)

        elif isinstance(dfs, list) or isinstance(dfs, dict) or isinstance(df, pd.DataFrame):
            all_dfs_info = {}
            if isinstance(df, pd.DataFrame):
                dfs = {'Sheet1': df}
            if isinstance(dfs, list):
                dfs = {f'Sheet{k}': v for k, v in enumerate(dfs, start=0)}
            workbook = openpyxl.Workbook()
            del workbook['Sheet']
            # for sheet_name, df in dfs.items():
            #     worksheet = workbook.create_sheet(sheet_name)
            #     insert_dataframe_to_worksheet(worksheet=worksheet, df=df)
            for sheet_name,dfs_dict in dfs.items():
                worksheet = workbook.create_sheet(sheet_name)
                start = 0
                dfs_info = {}
                for i in range(len(dfs_dict)):
                    df_info_ori=dfs_dict[i]
                    df = df_info_ori['dataframe']
                    df_name = df_info_ori['key']
                    empty_lines = df_info_ori.get('empty_lines',2)
                    m,n = df.shape
                    if 'title' in df_info_ori and df_info_ori['title'] is not None and df_info_ori['title'] != '':
                        d = parse_titles(df_info_ori['title'])
                    else:
                        d = parse_titles(df_name)

                    start = start + len(d)
                    dfs_info[df_name] = [m,n,start+1,len(d)]
                    insert_dataframes_to_worksheet(worksheet=worksheet,df_name=df_name,df=df,start=start,head=d,df_info_ori=df_info_ori)
                    start = start + m + empty_lines + 1
                all_dfs_info[sheet_name] = dfs_info
            

        
        self.row1_as_col = row1_as_col
        self.workbook = workbook
        self.sheetnames = self.workbook.sheetnames
        self.dfs_info = all_dfs_info

        print('ExcelModifier()')
        print(f'self.row1_as_col = {row1_as_col}')
        print(f'self.workbook = {workbook}')
        print(f'self.sheetnames = {self.workbook.sheetnames}')
        print(f"self.dfs_info = {self.dfs_info}")

        return None
    

    def __repr__(self) -> str:
        string = f'sheetnames = {self.sheetnames}'
        return string
    

    def __str__(self) -> str:
        return self.__repr__()
    
    
    def __getitem__(self, sheet_name):
        print(f'sheet_name = {sheet_name}')
        if isinstance(sheet_name, str):
            ret = SheetModifier(
                worksheet={sheet_name:self.workbook[sheet_name]},
                dfs_info= self.dfs_info, 
                row1_as_col=self.row1_as_col)
        elif isinstance(sheet_name, list):
            worksheet = {}
            for i in sheet_name:
                worksheet[i] = self.workbook[i]
            ret = SheetModifier(
                worksheet=worksheet,
                dfs_info= self.dfs_info, 
                row1_as_col=self.row1_as_col)
        elif isinstance(sheet_name, slice):
            worksheet = {}
            for i in self.dfs_info:
                worksheet[i] = self.workbook[i]
            ret = SheetModifier(
                worksheet=worksheet,
                dfs_info= self.dfs_info, 
                row1_as_col=self.row1_as_col)
        return ret
    

    def set_sheets_bolds(self, sheetnames = None, **kwargs):
        print('set_sheets_bolds()')
        if sheetnames is None:
            sheetnames = self.sheetnames
        for sheet_name in sheetnames:
            self[sheet_name].set_bold_sereis(**kwargs)
        return None
    
    
    def set_sheets_alignment(self, sheetnames = None, **kwargs):
        print('set_sheets_alignment()')
        if sheetnames is None:
            sheetnames = self.sheetnames
        for sheet_name in sheetnames:
            self[sheet_name].set_alignment(**kwargs)
        return None
    

    def set_sheets_filter(self, sheet_names = None, **kwargs):
        """
        save()

        :param freeze_panes: freeze_panes.
        :type param1: str

        :return: return None.
        :rtype: NoneType

        :raises ValueError: ValueError
        """
        print('set_sheets_filter()')
        if sheet_names is None:
            sheet_names = self.sheetnames
        for sheet_name in sheet_names:
            self[sheet_name].set_filter(**kwargs)
        return None
    

    def save(self, filename, border=True):
        """
        save()

        :param filename: path to file.
        :type param1: str

        :return: return None.
        :rtype: NoneType

        :raises ValueError: ValueError
        """
        # 遍历所有 sheet 并删除第二列
        for sheet in self.workbook.worksheets:
            # sheet.delete_cols(2)  # 删除第二列（列索引从 1 开始）
            if border:
                sheet.sheet_view.showGridLines = False  # 隐藏网格线
        print(f'save(filename = \'{filename}\')')
        self.workbook.save(filename=filename)
        return None
    
